//
//	pcx.cpp - source file / freeware
//
//	David Henry - tfc_duke@club-internet.fr
//


#ifdef _WIN32
#define	WIN32_LEAN_AND_MEAN
#include	<windows.h>
#endif

#include	<GL/gl.h>
#include	<fstream>

#include	"texture.h"


// --------------------------------------------
// PCXHEADER - pcx header structure.
// --------------------------------------------

#pragma pack(1)
typedef struct tagPCXHEADER
{
	unsigned char	manufacturer;		// fabriquant
	unsigned char	version;			// version
	unsigned char	encoding;			// encodage
	unsigned char	bitsPerPixel;		// nombre de bits par
										//  pixel pour un plan
	unsigned short	x, y;
	unsigned short	width, height;		// dimensions
	unsigned short	horzRes, vertRes;

	unsigned char	palette[48];		// palette couleur
	unsigned char	reserved;			// reserv
	unsigned char	numColorPlanes;		// nombre de plans

	unsigned short	bytesPerScanLine;	// octets par ligne
	unsigned short	paletteType;
	unsigned short	horzSize, vertSize;

	unsigned char	padding[54];

} PCXHEADER, *PPCXHEADER;
#pragma pack(4)



#ifndef WIN32
// --------------------------------------------
// RGBTRIPLE - pixel 24 bits.
// --------------------------------------------

typedef struct tagRGBTRIPLE			// rgbt
{
	unsigned char	rgbtBlue;		// bleu
	unsigned char	rgbtGreen;		// vert
	unsigned char	rgbtRed;		// rouge

} RGBTRIPLE, *PRGBTRIPLE;
#endif	// WIN32



// --------------------------------------------------
// LoadFilePCX() - charge une image PCX. Si bFlipVert
// vaut 'true', l'image est retourne verticalement.
// formats accepts:
//     # RGB 24 bits / version 5
//     # RLE  8 bits / version 5
//     # RLE  4 bits / version 2
//     # RLE  1 bit  / version 0
// --------------------------------------------------

bool LoadFilePCX( const std::string &szFileName, gl_texture_t **img, bool bFlipVert )
{
	std::ifstream	file;				// fichier
	PCXHEADER		*pcxh = NULL;		// header PCX
	RGBTRIPLE		*palette = NULL;	// pointeur palette couleur
	char			*buffer = NULL;		// buffer stockant l'intgralit du fichier
	GLubyte			*ptr = NULL;		// pointeur donnes pixels
	unsigned char	rle_count;			// 1er octet packet RLE
	unsigned char	rle_value;			// 2me octet packet RLE
	int				bitcount;			// nombre de bits par pixel


	/////////////////////////////////////////////////////
	// lit l'intgralit du fichier dans le buffer

	file.open( szFileName.c_str(), std::ios::in | std::ios::binary );

	if( file.fail() )
		return false;

	file.seekg( 0, std::ios::end );
	long flen = file.tellg();
	file.seekg( 0, std::ios::beg );

	buffer = new char[ flen ];
	file.read( buffer, flen );
	char *pBuff = buffer;

	file.close();

	/////////////////////////////////////////////////////

	pcxh = (PCXHEADER *)pBuff;
	pBuff = (char *)&buffer[128];


	// vrification du numro magique
	if( pcxh->manufacturer != 10 )
	{
		delete [] buffer;
		return false;
	}

	// calcul les dimensions relles de l'image
	pcxh->width	 = pcxh->width  - pcxh->x + 1;
	pcxh->height = pcxh->height - pcxh->y + 1;

	// on vrifie que ces dimensions sont valides
	if( (pcxh->width <= 0) || (pcxh->height <= 0) )
	{
		delete [] buffer;
		return false;
	}

	// cre un objet texture
	*img = new gl_texture_t;
	(*img)->width = pcxh->width;
	(*img)->height = pcxh->height;
	(*img)->format = GL_RGB;
	(*img)->internalFormat = 3;

	// allocatation mmoire pour les donnes pixels 32 bits
	(*img)->texels = new GLubyte[ (*img)->width * (*img)->height * (*img)->internalFormat ];
	ptr = &(*img)->texels[0];


	// initialisation du packet RLE  0
	rle_count = 0;
	rle_value = 0;

	// calcul du nombre rel de bits par pixel
	bitcount = pcxh->bitsPerPixel * pcxh->numColorPlanes;


	/////////////////////////////////////////////////////
	// lecture des donnes suivant le nombre de bits par
	// pixel.
	/////////////////////////////////////////////////////

	switch( bitcount )
	{
		case 1:
		{
// COLOR-MAPPED BLACK & WHITE 1 BIT
			// la palette se trouve dans l'en-tte du fichier
			palette = (RGBTRIPLE *)&pcxh->palette;

			for( int y = 0; y < (int)((*img)->height); y++ )
			{
				if( !bFlipVert ) {
					ptr = &(*img)->texels[ ((*img)->height - (y+1)) * (*img)->width * 3 ];
				}

				int bytes = pcxh->bytesPerScanLine;

				// lecture dcodage de la ligne y
				while( bytes-- )
				{
					if( rle_count == 0 )
					{
						if( (rle_value = *(pBuff++)) < 0xc0 )
						{
							rle_count = 1;
						}
						else
						{
							rle_count = rle_value - 0xc0;
							rle_value = *(pBuff++);
						}
					}

					rle_count--;

					for( int i = 7; i >= 0; i--, ptr += 3 )
					{
						int clrIdx = ((rle_value & (1<<i)) > 0);

						ptr[0] = (GLubyte)palette[ clrIdx ].rgbtBlue;
						ptr[1] = (GLubyte)palette[ clrIdx ].rgbtGreen;
						ptr[2] = (GLubyte)palette[ clrIdx ].rgbtRed;
					}
				}
			}

			break;
		}

		case 4:
		{
// COLOR-MAPPED BGR 4 BITS
			// la palette se trouve dans l'en-tte du fichier
			palette = (RGBTRIPLE *)&pcxh->palette;

			// indices couleur palette pour chaque pixel d'une ligne de l'image
			unsigned char *colorIdx = new unsigned char[ (*img)->width ];
			unsigned char *line = new unsigned char[ pcxh->bytesPerScanLine ];

			for( int y = 0; y < (int)((*img)->height); y++ )
			{
				if( !bFlipVert ) {
					ptr = &(*img)->texels[ ((*img)->height - (y+1)) * (*img)->width * 3 ];
				}
				else {
					ptr = &(*img)->texels[ y * (*img)->width * 3 ];
				}


				// remise  zro des indices couleur palette
				memset( colorIdx, 0, (*img)->width * sizeof( unsigned char ) );

				for( int c = 0; c < 4; c++ )
				{
					unsigned char *pLine = &line[0];
					int bytes = pcxh->bytesPerScanLine;

					// lecture et dcodage de la ligne y
					while( bytes-- )
					{
						if( rle_count == 0 )
						{
							if( (rle_value = *(pBuff++)) < 0xc0 )
							{
								rle_count = 1;
							}
							else
							{
								rle_count = rle_value - 0xc0;
								rle_value = *(pBuff++);
							}
						}

						rle_count--;
						*(pLine++) = rle_value;
					}

					// calcule les indices couleur palette de la ligne
					for( int x = 0; x < (int)((*img)->width); x++ )
					{
						if( line[ x / 8 ] & (128 >> (x % 8)) )
							colorIdx[x] += (1 << c);
					}
				}

				// criture de la ligne de pixels
				for( int x = 0; x < (int)((*img)->width); x++, ptr += 3 )
				{
					ptr[0] = (GLubyte)palette[ colorIdx[x] ].rgbtBlue;
					ptr[1] = (GLubyte)palette[ colorIdx[x] ].rgbtGreen;
					ptr[2] = (GLubyte)palette[ colorIdx[x] ].rgbtRed;
				}
			}

			delete [] colorIdx;
			delete [] line;

			break;
		}

		case 8:
		{
// COLOR-MAPPED BGR 8 BITS
			// la palette se trouve au 769 dernier octet du fichier
			char *ptrPal = &buffer[ flen - 769 ];

			// on vrifie la palette ; le premier char doit tre gal  12
			if( *(ptrPal++) != 12 )
			{
				delete [] buffer;
				delete *img;
				return false;
			}

			palette = (RGBTRIPLE *)ptrPal;

			for( int y = 0; y < (int)((*img)->height); y++ )
			{
				if( !bFlipVert ) {
					ptr = &(*img)->texels[ ((*img)->height - (y+1)) * (*img)->width * 3 ];
				}

				int bytes = pcxh->bytesPerScanLine;

				// lecture et dcodage de la ligne y
				while( bytes-- )
				{
					if( rle_count == 0 )
					{
						if( (rle_value = *(pBuff++)) < 0xc0 )
						{
							rle_count = 1;
						}
						else
						{
							rle_count = rle_value - 0xc0;
							rle_value = *(pBuff++);
						}
					}

					rle_count--;
					ptr[0] = (GLubyte)palette[ rle_value ].rgbtBlue;
					ptr[1] = (GLubyte)palette[ rle_value ].rgbtGreen;
					ptr[2] = (GLubyte)palette[ rle_value ].rgbtRed;
					ptr += 3;
				}
			}

			break;
		}

		case 24:
		{
// TRUE-COLOR RGB 24 BITS
			for( int y = (*img)->height - 1; y >= 0; y-- )
			{
				for( int c = 0; c < 3; c++ )
				{
					if( bFlipVert ) {
						ptr = &(*img)->texels[ ((*img)->height - (y+1)) * (*img)->width * 3 ];
					}
					else {
						ptr = &(*img)->texels[ y * (*img)->width * 3 ];
					}

					int bytes = pcxh->bytesPerScanLine;

					// lecture et dcodage de la ligne y
					while( bytes-- )
					{
						if( rle_count == 0 )
						{
							if( (rle_value = *(pBuff++)) < 0xc0 )
							{
								rle_count = 1;
							}
							else
							{
								rle_count = rle_value - 0xc0;
								rle_value = *(pBuff++);
							}
						}

						rle_count--;
						ptr[c] = (GLubyte)rle_value;
						ptr += 3;
					}
				}
			}

			break;
		}

		default:
		{
			// format inconnu ou non trat
			delete [] buffer;
			delete *img;
			return false;
		}
	}

	delete [] buffer;
	return true;
}
