//
//	texture.cpp - source file
//
//	David Henry - tfc_duke@club-internet.fr
//


#ifdef _WIN32
#define	WIN32_LEAN_AND_MEAN
#include	<windows.h>
#endif

#include	<GL/gl.h>
#include	<GL/glu.h>
#include	<iostream>

#include	"texture.h"



/////////////////////////////////////////////////
//				CTextureManager				   //
/////////////////////////////////////////////////


// ----------------------------------------------
// Initialize() - initialise le manager. Cre une
// texture d'chiquier pour la texture 'default'.
// ----------------------------------------------

bool CTextureManager::Initialize( void )
{
	// on vide la base de donnes
	ReleaseTextures();

	// c'est la premire texture charge en mmoire. Si
	// une texture ne peut tre lue ou charge, on utilise
	// celle si  la place.
	m_pDefaultTex = new gl_texture_t;
	if( !m_pDefaultTex )
		return false;

	// cration et initialisation de la texture 'default'
	if( CreateDefaultTexture( 64 ) )
	{
		glGenTextures( 1, &m_pDefaultTex->id );
		CreateGLTexture( m_pDefaultTex );

		// OpenGL possde sa propre copie de l'image
		delete [] m_pDefaultTex->texels;
		m_pDefaultTex->texels = NULL;

		// ajout  la base de donne
		m_kTextures["default"] = m_pDefaultTex;
		m_iTextureCount++;

		std::cout << "CTextureManager : creating default texture, id [" << m_pDefaultTex->id << "]" << std::endl;
		return true;
	}

	return false;
}



// ----------------------------------------------
// ReleaseTextures() - dtruit toutes les textures
// stockes dans la base de donnes.
// ----------------------------------------------

void CTextureManager::ReleaseTextures( void )
{
	// on dtruit chaque texture de la base de donnes du manager
	for( m_kItor = m_kTextures.begin(); m_kItor != m_kTextures.end(); )
	{
		// dtruit la texture OpenGL
		glDeleteTextures( 1, &(*m_kItor).second->id );

		std::cout << "CTextureManager : deleting " << (*m_kItor).first << " texture" << std::endl;

		// dtruit l'objet texture
		delete (*m_kItor).second;
		m_kTextures.erase( m_kItor++ );
	}

	m_iTextureCount = 0;
}



// ----------------------------------------------
// LoadTextureFromFile() - charge une texture
// depuis un fichier image.
// ----------------------------------------------

bool CTextureManager::LoadTextureFromFile( const std::string &szFileName, gl_texture_t **img, bool bFlipVert )
{
	if( strstr( szFileName.c_str(), ".png" ) || strstr( szFileName.c_str(), ".PNG" ) )
	{
		return LoadFilePNG( szFileName, img, bFlipVert );
	}
	else if( strstr( szFileName.c_str(), ".jpg" ) || strstr( szFileName.c_str(), ".JPG" ) )
	{
		return LoadFileJPEG( szFileName, img, bFlipVert );
	}
	else if( strstr( szFileName.c_str(), ".tga" ) || strstr( szFileName.c_str(), ".TGA" ) )
	{
		return LoadFileTGA( szFileName, img, bFlipVert );
	}
	else if( strstr( szFileName.c_str(), ".bmp" ) || strstr( szFileName.c_str(), ".BMP" ) )
	{
		return LoadFileBMP( szFileName, img, bFlipVert );
	}
	else if( strstr( szFileName.c_str(), ".pcx" ) || strstr( szFileName.c_str(), ".PCX" ) )
	{
		return LoadFilePCX( szFileName, img, bFlipVert );
	}
	else
	{
		// format inconnu
		return false;
	}
}



// ----------------------------------------------
// LoadTexture() - charge une texture. Vrifie
// si elle n'est pas dj stocke en mmoire. Si
// c'est le cas, la fonction retourne l'ID de la
// texture dj stocke. Sinon, elle retourne
// l'ID de la nouvelle texture.
// ----------------------------------------------

GLuint CTextureManager::LoadTexture( const std::string &szFileName, bool bFlipVert /* = false */ )
{
	// on tente d'obtenir la texture depuis la base de donnes.
	// Si elle a dj t charge au pralable, on retourne son
	// ID, sinon l'ID de l'objet retourn est celle de la texture
	// par dfaut. On tente alors de la charger dans la base.
	gl_texture_t *img = GetTexture( szFileName );

	if( IsDefaultTexture( img ) )
	{
		// on tente de charger la texture
		if( LoadTextureFromFile( szFileName, &img, bFlipVert ) )
		{
			// cration et initialisation de la texture opengl
			glGenTextures( 1, &img->id );
			CreateGLTexture( img );

			// dsallocation de mmoire - OpenGL possde sa propre copie de l'image
			delete [] img->texels;
			img->texels = NULL;

			// cration d'une nouvelle texture et ajout  la base de donnes
			m_kTextures[szFileName] = img;
			m_iTextureCount++;
		}
		else
		{
			// impossible de charger la texture, texture par dfault utilise
			std::cout << "CTextureManager : failed to load " << szFileName << "!"
					  << "giving default texture id [" << m_pDefaultTex->id << "]" << std::endl;
			img = m_pDefaultTex;
		}
	}

	std::cout << "CTextureManager : " << szFileName << ", id [" << img->id << "]" << std::endl;
	return img->id;
}



// ----------------------------------------------
// CreateTexture() - cre une texture OpenGL 
// partir d'un objet gl_texture_t.
// ----------------------------------------------

void CTextureManager::CreateGLTexture( gl_texture_t *img )
{
	glBindTexture( GL_TEXTURE_2D, img->id );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

//	glTexImage2D( GL_TEXTURE_2D, 0, img->internalFormat, img->width, img->height,
//					0, img->format, GL_UNSIGNED_BYTE, img->texels );
	gluBuild2DMipmaps( GL_TEXTURE_2D, img->format, img->width, img->height,
					img->format, GL_UNSIGNED_BYTE, img->texels );
}



// ----------------------------------------------
// CreateDefaultTexture() - cre la texture
// 'default' qui sera utilise lorsque la recherche
// ou le chargement d'une texture a chou.
// ----------------------------------------------

bool CTextureManager::CreateDefaultTexture( unsigned int size )
{
	if( !m_pDefaultTex )
		return false;

	// cration d'une texture chiquier pour la texture par dfaut
	unsigned int i, j, c;	// variables de contrle

	m_pDefaultTex->width = size;
	m_pDefaultTex->height = size;
	m_pDefaultTex->format = GL_LUMINANCE;
	m_pDefaultTex->internalFormat = 1;

	unsigned int iSize = m_pDefaultTex->width * m_pDefaultTex->height
							* m_pDefaultTex->internalFormat;

	m_pDefaultTex->texels = new GLubyte[ iSize ];
	if( !m_pDefaultTex->texels )
		return false;

	// construction d'une texture d'chiquier
	for( i = 0; i < m_pDefaultTex->height; i++ )
	{
		for( j = 0; j < m_pDefaultTex->width; j++ )
		{
			c = ( !(i & 8) ^ !(j & 8)) * 255;
			m_pDefaultTex->texels[ (i * m_pDefaultTex->width) + j ] = (GLubyte)c;
		}
	}

	return true;
}



// ----------------------------------------------
// GetTexture() - recherche une texture dans la
// base de donne du manager. Si elle n'est pas
// trouve, on retourne la texture 'default'.
// ----------------------------------------------

gl_texture_t *CTextureManager::GetTexture( const std::string &szFileName )
{
	if( (m_kItor = m_kTextures.find( szFileName )) != m_kTextures.end() )
	{
		return (*m_kItor).second;
	}
	else
	{
		return m_pDefaultTex;
	}
}
